@@ -1,5 +1,6 @@
 """Semantic analysis of types"""
 
+from abc import abstractmethod
 from collections import OrderedDict
 from typing import Callable, List, Optional, Set, Tuple, Iterator, TypeVar, Iterable, Dict, Union
 
@@ -23,11 +24,12 @@ from mypy.nodes import (
     TVAR, TYPE_ALIAS, UNBOUND_IMPORTED, TypeInfo, Context, SymbolTableNode, Var, Expression,
     IndexExpr, RefExpr, nongen_builtins, check_arg_names, check_arg_kinds, ARG_POS, ARG_NAMED,
     ARG_OPT, ARG_NAMED_OPT, ARG_STAR, ARG_STAR2, TypeVarExpr, FuncDef, CallExpr, NameExpr,
-    Decorator, Node
+    Decorator, Node, ImportedName, type_aliases
 )
 from mypy.tvar_scope import TypeVarScope
 from mypy.exprtotype import expr_to_unanalyzed_type, TypeTranslationError
 from mypy.plugin import Plugin, TypeAnalyzerPluginInterface, AnalyzeTypeContext
+from mypy.semanal_shared import SemanticAnalyzerCoreInterface
 from mypy import nodes, messages
 
 
@@ -53,11 +55,8 @@ ARG_KINDS_BY_CONSTRUCTOR = {
 
 
 def analyze_type_alias(node: Expression,
-                       lookup_func: Callable[[str, Context], Optional[SymbolTableNode]],
-                       lookup_fqn_func: Callable[[str], SymbolTableNode],
+                       api: SemanticAnalyzerCoreInterface,
                        tvar_scope: TypeVarScope,
-                       fail_func: Callable[[str, Context], None],
-                       note_func: Callable[[str, Context], None],
                        plugin: Plugin,
                        options: Options,
                        is_typeshed_stub: bool,
@@ -79,7 +78,7 @@ def analyze_type_alias(node: Expression,
         # class-referenced type variable as a type alias.  It's easier to catch
         # that one in checkmember.py
         if node.kind == TVAR:
-            fail_func('Type variable "{}" is invalid as target for type alias'.format(
+            api.fail('Type variable "{}" is invalid as target for type alias'.format(
                 node.fullname), node)
             return None
         if not (isinstance(node.node, TypeInfo) or
@@ -102,8 +101,8 @@ def analyze_type_alias(node: Expression,
     elif isinstance(node, CallExpr):
         if (isinstance(node.callee, NameExpr) and len(node.args) == 1 and
                 isinstance(node.args[0], NameExpr)):
-            call = lookup_func(node.callee.name, node.callee)
-            arg = lookup_func(node.args[0].name, node.args[0])
+            call = api.lookup_qualified(node.callee.name, node.callee)
+            arg = api.lookup_qualified(node.args[0].name, node.args[0])
             if (call is not None and call.node and call.node.fullname() == 'builtins.type' and
                     arg is not None and arg.node and arg.node.fullname() == 'builtins.None'):
                 return NoneTyp(), set()
@@ -116,10 +115,9 @@ def analyze_type_alias(node: Expression,
     try:
         type = expr_to_unanalyzed_type(node)
     except TypeTranslationError:
-        fail_func('Invalid type alias', node)
+        api.fail('Invalid type alias', node)
         return None
-    analyzer = TypeAnalyser(lookup_func, lookup_fqn_func, tvar_scope, fail_func, note_func,
-                            plugin, options, is_typeshed_stub, aliasing=True,
+    analyzer = TypeAnalyser(api, tvar_scope, plugin, options, is_typeshed_stub, aliasing=True,
                             allow_unnormalized=allow_unnormalized, warn_bound_tvar=warn_bound_tvar)
     analyzer.in_dynamic_func = in_dynamic_func
     analyzer.global_scope = global_scope
@@ -147,11 +145,8 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
     global_scope = True  # type: bool
 
     def __init__(self,
-                 lookup_func: Callable[[str, Context], Optional[SymbolTableNode]],
-                 lookup_fqn_func: Callable[[str], SymbolTableNode],
+                 api: SemanticAnalyzerCoreInterface,
                  tvar_scope: Optional[TypeVarScope],
-                 fail_func: Callable[[str, Context], None],
-                 note_func: Callable[[str, Context], None],
                  plugin: Plugin,
                  options: Options,
                  is_typeshed_stub: bool, *,
@@ -160,10 +155,11 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
                  allow_unnormalized: bool = False,
                  third_pass: bool = False,
                  warn_bound_tvar: bool = False) -> None:
-        self.lookup = lookup_func
-        self.lookup_fqn_func = lookup_fqn_func
-        self.fail_func = fail_func
-        self.note_func = note_func
+        self.api = api
+        self.lookup = api.lookup_qualified
+        self.lookup_fqn_func = api.lookup_fully_qualified
+        self.fail_func = api.fail
+        self.note_func = api.note
         self.tvar_scope = tvar_scope
         self.aliasing = aliasing
         self.allow_tuple_literal = allow_tuple_literal
@@ -179,13 +175,37 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
         self.aliases_used = set()  # type: Set[str]
 
     def visit_unbound_type(self, t: UnboundType) -> Type:
+        typ = self.visit_unbound_type_nonoptional(t)
         if t.optional:
-            t.optional = False
             # We don't need to worry about double-wrapping Optionals or
             # wrapping Anys: Union simplification will take care of that.
-            return make_optional_type(self.visit_unbound_type(t))
-        sym = self.lookup(t.name, t, suppress_errors=self.third_pass)  # type: ignore
+            return make_optional_type(typ)
+        return typ
+
+    def visit_unbound_type_nonoptional(self, t: UnboundType) -> Type:
+        sym = self.lookup(t.name, t, suppress_errors=self.third_pass)
+        if '.' in t.name:
+            # Handle indirect references to imported names.
+            #
+            # TODO: Do this for module-local references as well and remove ImportedName
+            #    type check below.
+            sym = self.api.dereference_module_cross_ref(sym)
         if sym is not None:
+            if isinstance(sym.node, ImportedName):
+                # Forward reference to an imported name that hasn't been processed yet.
+                # To maintain backward compatibility, these get translated to Any.
+                #
+                # TODO: Remove this special case.
+                return AnyType(TypeOfAny.implementation_artifact)
+            if sym.fullname in type_aliases:
+                # Resolve forward reference to type alias like 'typing.List'.
+                # TODO: Unify how type aliases are handled; currently we resolve them in two
+                #     places (the other is in the semantic analyzer pass 2).
+                resolved = type_aliases[sym.fullname]
+                new = self.api.lookup_qualified(resolved, t)
+                if new:
+                    sym = new.copy()
+                    sym.normalized = True
             if sym.node is None:
                 # UNBOUND_IMPORTED can happen if an unknown name was imported.
                 if sym.kind != UNBOUND_IMPORTED:
@@ -298,7 +318,8 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
                     # context. This is slightly problematic as it allows using the type 'Any'
                     # as a base class -- however, this will fail soon at runtime so the problem
                     # is pretty minor.
-                    return AnyType(TypeOfAny.from_unimported_type)
+                    return AnyType(TypeOfAny.from_unimported_type,
+                                   missing_import_name=sym.node.type.missing_import_name)
                 # Allow unbound type variables when defining an alias
                 if not (self.aliasing and sym.kind == TVAR and
                         (not self.tvar_scope or self.tvar_scope.get_binding(sym) is None)):
@@ -398,9 +419,11 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
         # Types such as (t1, t2, ...) only allowed in assignment statements. They'll
         # generate errors elsewhere, and Tuple[t1, t2, ...] must be used instead.
         if t.implicit and not self.allow_tuple_literal:
-            self.fail('Invalid tuple literal type', t)
+            self.fail('Syntax error in type annotation', t)
             if len(t.items) == 1:
                 self.note_func('Suggestion: Is there a spurious trailing comma?', t)
+            else:
+                self.note_func('Suggestion: Use Tuple[T1, ..., Tn] instead of (T1, ..., Tn)', t)
             return AnyType(TypeOfAny.from_error)
         star_count = sum(1 for item in t.items if isinstance(item, StarType))
         if star_count > 1:
@@ -659,19 +682,17 @@ class TypeAnalyserPass3(TypeVisitor[None]):
     """
 
     def __init__(self,
-                 lookup_func: Callable[[str, Context], Optional[SymbolTableNode]],
-                 lookup_fqn_func: Callable[[str], SymbolTableNode],
-                 fail_func: Callable[[str, Context], None],
-                 note_func: Callable[[str, Context], None],
+                 api: SemanticAnalyzerCoreInterface,
                  plugin: Plugin,
                  options: Options,
                  is_typeshed_stub: bool,
                  indicator: Dict[str, bool],
                  patches: List[Tuple[int, Callable[[], None]]]) -> None:
-        self.lookup_func = lookup_func
-        self.lookup_fqn_func = lookup_fqn_func
-        self.fail = fail_func
-        self.note_func = note_func
+        self.api = api
+        self.lookup_func = api.lookup_qualified
+        self.lookup_fqn_func = api.lookup_fully_qualified
+        self.fail = api.fail
+        self.note_func = api.note
         self.options = options
         self.plugin = plugin
         self.is_typeshed_stub = is_typeshed_stub
@@ -799,11 +820,8 @@ class TypeAnalyserPass3(TypeVisitor[None]):
             t.resolve(resolved)
 
     def anal_type(self, tp: UnboundType) -> Type:
-        tpan = TypeAnalyser(self.lookup_func,
-                            self.lookup_fqn_func,
+        tpan = TypeAnalyser(self.api,
                             None,
-                            self.fail,
-                            self.note_func,
                             self.plugin,
                             self.options,
                             self.is_typeshed_stub,
